//@ts-check
"use strict";

/**
 * @enum {number}
 */
const BaddieStates = {
  Idle: 0,
  Wander: 1,
  ChaseTarget: 2,
  AttackTarget: 3,
  Hit: 4,
  RushToSpawn: 5
}

/**
@typedef BaddieSpawner
@type {{
  pos: PIXI.Point;
  baddieName: BaddieName;
  targetAmount: number;
  currentAmount: number;
}}
*/

/**
@typedef Baddie
@type {{
  baddieDesc: BaddieDesc;
  spawner: BaddieSpawner;
  pos: PIXI.Point;
  spawnPos: PIXI.Point;
  collideBox: { width: number; height: number; };
  hitbox: { width: number; height: number; };
  speed: number;
  health: number;

  aggressive: boolean;
  state: BaddieStates;
  stateTimer: number;
  stateTarget: EntityLike;
  knockbackDirection: PIXI.Point;
  beforeWanderDelay: number;
  wanderAngle: number;

  pixiRoot: PIXI.Container;
  bodySpritesheet: Spritesheet;
  bodySprite: PIXI.AnimatedSprite;
  shadowSprite: PIXI.Sprite;
}}
*/

/**
 * @typedef BaddieName
 * @type {keyof typeof baddieDescs}
 */

/**
@typedef BaddieDesc
@type {{
  id: BaddieName;
  maxHealth: number;
  attackDamage: number;
  speed: number;
  aggroRange: number;
  hitStunDelay: number;
  knockbackSpeed: number;

  textureVariants?: number;
  flying?: boolean;
  dropItem?: ItemName;
}}
 */

/**
@type {{
  bug: BaddieDesc,
  champi: BaddieDesc,
  oiecorne: BaddieDesc,
  moutoncorne: BaddieDesc,
  topicorne: BaddieDesc,
  topacorne: BaddieDesc,
  kamoulox: BaddieDesc,
  poulpicorne: BaddieDesc,
  bat: BaddieDesc,
  albinosbat: BaddieDesc,
  blackdragon: BaddieDesc,
  yellowdragon: BaddieDesc,
  champiking: BaddieDesc,
  petitmouton: BaddieDesc,
  flower: BaddieDesc,
}}
*/
const baddieDescs = {
  bug: {
    id: "bug",
    maxHealth: 4,
    attackDamage: 1,
    speed: 3,
    aggroRange: 4,
    hitStunDelay: 0.3,
    knockbackSpeed: 0.05
  },
  champi: {
    id: "champi",
    maxHealth: 5,
    attackDamage: 1,
    speed: 2,
    aggroRange: 4,
    hitStunDelay: 0.3,
    knockbackSpeed: 0.05,
    textureVariants: 4
  },
  oiecorne: {
    id: "oiecorne",
    maxHealth: 10,
    attackDamage: 1,
    speed: 4.5,
    aggroRange: 4,
    hitStunDelay: 0.3,
    knockbackSpeed: 0.05,
    textureVariants: 2
  },
  moutoncorne: {
    id: "moutoncorne",
    maxHealth: 25,
    attackDamage: 1,
    speed: 4,
    aggroRange: 4,
    hitStunDelay: 0.0,
    knockbackSpeed: 0.0,
    textureVariants: 2
  },
  topicorne: {
    id: "topicorne",
    maxHealth: 6,
    attackDamage: 1,
    speed: 3,
    aggroRange: 4,
    hitStunDelay: 0.3,
    knockbackSpeed: 0.05,
    textureVariants: 1
  },
  topacorne: {
    id: "topacorne",
    maxHealth: 10,
    attackDamage: 1,
    speed: 3,
    aggroRange: 4,
    hitStunDelay: 0.3,
    knockbackSpeed: 0.05,
    textureVariants: 2
  },
  kamoulox: {
    id: "kamoulox",
    maxHealth: 12,
    attackDamage: 2,
    speed: 4,
    aggroRange: 4,
    hitStunDelay: 0.2,
    knockbackSpeed: 0.03,
    textureVariants: 1
  },
  poulpicorne: {
    id: "poulpicorne",
    maxHealth: 5,
    attackDamage: 1,
    speed: 0,
    aggroRange: 4,
    hitStunDelay: 0.3,
    knockbackSpeed: 0,
    textureVariants: 1
  },
  bat: {
    id: "bat",
    maxHealth: 6,
    attackDamage: 1,
    speed: 3.5,
    aggroRange: 4,
    hitStunDelay: 0.3,
    knockbackSpeed: 0.05,
    textureVariants: 1,
    flying: true
  },
  albinosbat: {
    id: "albinosbat",
    maxHealth: 100,
    attackDamage: 2,
    speed: 5,
    aggroRange: 4,
    hitStunDelay: 0.3,
    knockbackSpeed: 0.05,
    textureVariants: 1,
    flying: true,
    dropItem: "ham"
  },
  blackdragon: {
    id: "blackdragon",
    maxHealth: 120,
    attackDamage: 1,
    speed: 4,
    aggroRange: 5,
    hitStunDelay: 0.0,
    knockbackSpeed: 0.0,
    textureVariants: 1,
  },
  yellowdragon: {
    id: "yellowdragon",
    maxHealth: 80,
    attackDamage: 1,
    speed: 3.5,
    aggroRange: 5,
    hitStunDelay: 0.0,
    knockbackSpeed: 0.0,
    textureVariants: 1,
  },
  champiking: {
    id: "champiking",
    maxHealth: 25,
    attackDamage: 2,
    speed: 3,
    aggroRange: 4,
    hitStunDelay: 0.3,
    knockbackSpeed: 0.05,
    textureVariants: 1
  },
  petitmouton: {
    id: "petitmouton",
    maxHealth: 5,
    attackDamage: 1,
    speed: 3.5,
    aggroRange: 0,
    hitStunDelay: 0.2,
    knockbackSpeed: 0.04,
    textureVariants: 2
  },
  flower: {
    id: "flower",
    maxHealth: 10,
    attackDamage: 1,
    speed: 3,
    aggroRange: 4,
    hitStunDelay: 0.3,
    knockbackSpeed: 0.05,
    textureVariants: 1
  },
};

let combatCountdown = 0;

/**
 * @param {World} world
 * @param {BaddieSpawner} spawner
 * @param {BaddieDesc} baddieDesc
 * @returns Baddie
* */
function createBaddie(world, spawner, baddieDesc) {
  const pixiRoot = new PIXI.Container();
  pixiRoot.scale.set(inverseTileSize);
  world.pixiLayers["ground"].entitiesContainer.addChild(pixiRoot);

  const pos = spawner.pos.clone();
  pos.x += randFloat(-0.5, 0.5);
  pos.y += randFloat(-0.5, 0.5);
  pixiRoot.position.copyFrom(pos);

  const bodySpritesheet = getSpritesheet(`inGame/baddies/${baddieDesc.id}/body`, baddieDesc.textureVariants);
  const bodySprite = new PIXI.AnimatedSprite(bodySpritesheet.animations["idle"], false);
  bodySprite.anchor.copyFrom(bodySpritesheet.anchor);
  pixiRoot.addChild(bodySprite);

  const shadowSprite = new PIXI.Sprite(tex("inGame/shadow"));
  shadowSprite.scale.set(inverseTileSize * 1.2);
  shadowSprite.anchor.set(0.5, 0);
  world.pixiShadowsLayer.addChild(shadowSprite);

  /** @type Baddie */
  const baddie = {
    pos,
    spawnPos: spawner.pos,
    collideBox: { width: 0.8, height: 0.4 },
    hitbox: { width: 1.5, height: 2 },
    baddieDesc,
    spawner,
    speed: baddieDesc.speed,
    health: baddieDesc.maxHealth,

    aggressive: false,
    state: BaddieStates.Idle,
    stateTimer: 0,
    stateTarget: null,
    knockbackDirection: new PIXI.Point(),
    beforeWanderDelay: 0,
    wanderAngle: 0,

    pixiRoot,
    bodySpritesheet,
    bodySprite,
    shadowSprite
  };

  setBaddieIdle(baddie, 1);
  world.baddies.push(baddie);

  return baddie;
}


/** @param {number} dt */
function updateBaddieSpawners(dt, ignoreBaxtonPos = false) {
  for (const spawner of inGame.activeWorld.baddieSpawners) {
    const dist = distxy(spawner.pos, inGame.baxton.pos);
    if (spawner.currentAmount != spawner.targetAmount && (dist > chunkSize || ignoreBaxtonPos)) {
      for (let i = 0; i < spawner.targetAmount - spawner.currentAmount; i++) {
        createBaddie(inGame.activeWorld, spawner, baddieDescs[spawner.baddieName]);
      }

      spawner.targetAmount = spawner.currentAmount;
    }
  }
}

/** @param {number} dt */
function updateBaddies(dt) {
  const { baxton } = inGame;
  const { baddies, npcs } = inGame.activeWorld;

  let attackingBaddies = 0;

  for (const baddie of baddies) {
    if (inGame.gameOver) {
      if (baddie.state !== BaddieStates.Idle) setBaddieIdle(baddie, 1);
      continue;
    }

    /**
     * @type {EntityLike[]}
     */
    const availableTargets = [];
    for (let npc of npcs) {
      if (npc.following) availableTargets.push(npc);
    }

    availableTargets.push(baxton);

    let lookForTarget = () => {
      for (let target of availableTargets) {
        const diff = { x: target.pos.x - baddie.pos.x, y: target.pos.y - baddie.pos.y };
        const dist = lengthxy(diff);

        const chaseRange = baddie.aggressive ? 9 : baddie.baddieDesc.aggroRange;

        if (dist < chaseRange) {
          setBaddieChaseTarget(baddie, target);
          break;
        }
      }
    }

    switch (baddie.state) {
      case BaddieStates.Idle: {
        baddie.stateTimer -= dt;

        if (baddie.stateTimer <= 0) {
          lookForTarget();
          if (baddie.state === BaddieStates.ChaseTarget) break;

          baddie.aggressive = false;

          const distToBaxton = distxy(baddie.pos, baxton.pos);
          const distToSpawn = distxy(baddie.pos, baddie.spawnPos);
          if (distToSpawn > 3 && distToBaxton > chunkSize / 2) {
            setBaddieRushToSpawn(baddie);
            break;
          }

          if (-baddie.stateTimer > baddie.beforeWanderDelay && baddie.speed !== 0) {
            setBaddieWander(baddie);
          }
        }
        break;
      }
      case BaddieStates.Wander: {
        lookForTarget();
        if (baddie.state === BaddieStates.ChaseTarget) break;

        baddie.stateTimer -= dt;
        if (baddie.stateTimer <= 0) {
          setBaddieIdle(baddie, randFloat(1, 3));
        } else {
          const amount = 2 * dt;
          const move = { x: Math.cos(baddie.wanderAngle) * amount, y: Math.sin(baddie.wanderAngle) * amount }
          setBaddieDirection(baddie, move.x < 0);
          moveWithCollision(baddie, move, baddie.baddieDesc.flying === true);
        }

        break;
      }
      case BaddieStates.RushToSpawn: {
        const diff = { x: baddie.spawnPos.x - baddie.pos.x, y: baddie.spawnPos.y - baddie.pos.y };
        const dist = lengthxy(diff);

        if (dist <= 1) {
          setBaddieIdle(baddie, 1);
        } else {
          const rushSpeed = 10;
          const angle = Math.atan2(diff.y, diff.x);
          const amount = Math.min(dist, rushSpeed * dt);
          baddie.pos.x += Math.cos(angle) * amount;
          baddie.pos.y += Math.sin(angle) * amount;
        }

        break;
      }
      case BaddieStates.ChaseTarget: {
        attackingBaddies++;

        if (availableTargets.indexOf(baddie.stateTarget) === -1) {
          setBaddieIdle(baddie, 0.1);
          break;
        }

        baddie.stateTimer += dt;

        const diff = { x: baddie.stateTarget.pos.x - baddie.pos.x, y: baddie.stateTarget.pos.y - baddie.pos.y };
        const angle = Math.atan2(diff.y, diff.x);
        const dist = lengthxy(diff);

        const minChaseDelay = 2;
        const maxChaseDelay = 5;
        const stopChaseRange = 12;

        if (Math.abs(diff.x) < baddie.collideBox.width + baddie.stateTarget.hitbox.width * 0.9 &&
          Math.abs(diff.y) < baddie.collideBox.height + baddie.stateTarget.hitbox.height * 0.9) {

          setBaddieAttack(baddie);
        } else if (baddie.stateTimer > maxChaseDelay || (dist > stopChaseRange && baddie.stateTimer > minChaseDelay)) {
          setBaddieIdle(baddie, 1);
        } else {
          const amount = baddie.speed * dt;
          const move = { x: Math.cos(angle) * amount, y: Math.sin(angle) * amount }
          setBaddieDirection(baddie, move.x < 0);
          moveWithCollision(baddie, move, baddie.baddieDesc.flying === true);
        }

        break;
      }
      case BaddieStates.AttackTarget: {
        attackingBaddies++;

        baddie.stateTimer -= dt;
        if (baddie.stateTimer <= 0) {
          let attackLanded = false;

          if (availableTargets.indexOf(baddie.stateTarget) !== -1) {
            const diff = { x: baddie.stateTarget.pos.x - baddie.pos.x, y: baddie.stateTarget.pos.y - baddie.pos.y };

            if (Math.abs(diff.x) < baddie.collideBox.width + baddie.stateTarget.hitbox.width &&
              Math.abs(diff.y) < baddie.collideBox.height + baddie.stateTarget.hitbox.height) {

              attackLanded = true;
              const attackAngle = Math.atan2(-diff.y, -diff.x);

              if (baddie.stateTarget == baxton) {
                damageBaxton(1, attackAngle);
              }
              else {
                /** @type {any} */
                const npcTarget = baddie.stateTarget;
                damageNpc(npcTarget, baddie.baddieDesc.attackDamage, attackAngle);
              }
            }
          }

          setBaddieIdle(baddie, attackLanded ? 1 : 0.1);
        }
        break;
      }
      case BaddieStates.Hit: {
        baddie.stateTimer -= dt;
        if (baddie.stateTimer <= 0) {
          setBaddieIdle(baddie, baddie.baddieDesc.hitStunDelay);
        } else {
          moveWithCollision(baddie, baddie.knockbackDirection, baddie.baddieDesc.flying === true);
          baddie.knockbackDirection.x *= 0.9;
          baddie.knockbackDirection.y *= 0.9;
        }
        break;
      }
    }
  }

  // Combat music
  const isNowInCombat = attackingBaddies > 0;

  if (isNowInCombat) {
    combatCountdown = 5;

    if (!isInCombat) {
      isInCombat = true;
      updateCombatBgmVolume();
    }
  } else if (isInCombat) {
    combatCountdown -= dt;

    if (combatCountdown <= 0) {
      isInCombat = false;
      updateCombatBgmVolume();
    }
  }

  // Pushaway
  for (const baddie of baddies) {
    for (const otherEnemy of baddies) {
      if (baddie !== otherEnemy) pushaway(baddie, otherEnemy, dt);
    }

    pushaway(baddie, baxton, dt);
  }

  for (const baddie of baddies) {
    baddie.pixiRoot.position.copyFrom(baddie.pos);
    baddie.pixiRoot.zIndex = baddie.pos.y;

    baddie.shadowSprite.position.copyFrom(baddie.pos);
    baddie.shadowSprite.zIndex = baddie.pos.y - 10;

    baddie.bodySprite.update(app.ticker.deltaTime);
  }
}

/**
 *
 * @param {Baddie} baddie
 * @param {boolean} isLookingLeft
 */
function setBaddieDirection(baddie, isLookingLeft) {
  baddie.bodySprite.scale.x = isLookingLeft ? -1 : 1;
}

/**
 * @param {Baddie} baddie
 */
function setBaddieIdle(baddie, minIdleTime) {
  baddie.state = BaddieStates.Idle;
  baddie.stateTimer = minIdleTime;
  baddie.beforeWanderDelay = randFloat(0, 1.5);

  baddie.bodySprite.loop = true;
  setAnimation(baddie.bodySprite, baddie.bodySpritesheet.animations.idle);
}

/**
 * @param {Baddie} baddie
 */
function setBaddieWander(baddie) {
  baddie.state = BaddieStates.Wander;
  baddie.stateTimer = randFloat(0.5, 1);

  if (distxy(baddie.pos, baddie.spawnPos) >= 3) {
    baddie.wanderAngle = anglexy(baddie.spawnPos, baddie.pos) + randFloat(-Math.PI / 8, Math.PI / 8);
  } else {
    baddie.wanderAngle = randFloat(0, Math.PI * 2);
  }

  baddie.bodySprite.loop = true;
  setAnimation(baddie.bodySprite, baddie.bodySpritesheet.animations.run);
}

/**
 * @param {Baddie} baddie
 * @param {EntityLike} target
 */
function setBaddieChaseTarget(baddie, target) {
  setBaddieAggressive(baddie);
  baddie.state = BaddieStates.ChaseTarget;
  baddie.stateTarget = target;
  baddie.stateTimer = 0;

  baddie.bodySprite.loop = true;
  setAnimation(baddie.bodySprite, baddie.bodySpritesheet.animations.run);
}

/**
 * @param {Baddie} baddie
 */
function setBaddieAttack(baddie) {
  baddie.state = BaddieStates.AttackTarget;
  baddie.stateTimer = 0.3;

  baddie.bodySprite.loop = false;
  setAnimation(baddie.bodySprite, baddie.bodySpritesheet.animations.attack);
}

/**
 * @param {Baddie} baddie
 */
function setBaddieHit(baddie, angle) {
  setBaddieAggressive(baddie);
  baddie.state = BaddieStates.Hit;
  baddie.stateTimer = 0.3;

  baddie.knockbackDirection.set(-Math.cos(angle) * baddie.baddieDesc.knockbackSpeed, -Math.sin(angle) * baddie.baddieDesc.knockbackSpeed);

  baddie.bodySprite.loop = false;
  setAnimation(baddie.bodySprite, baddie.bodySpritesheet.animations.hit);
}

/**
 * @param {Baddie} baddie
 */
function setBaddieRushToSpawn(baddie) {
  baddie.aggressive = false;
  baddie.state = BaddieStates.RushToSpawn;
  baddie.health = baddie.baddieDesc.maxHealth;

  baddie.bodySprite.loop = true;
  setAnimation(baddie.bodySprite, baddie.bodySpritesheet.animations.run);
}

function setBaddieAggressive(baddie) {
  baddie.aggressive = true;

  for (let otherBaddie of inGame.activeWorld.baddies) {
    if (distxy(baddie.pos, otherBaddie.pos) <= 6) otherBaddie.aggressive = true;
  }
}

/**
 * @param {Baddie} baddie
 * @param {number} damage
 * @param {number} angle
 */
function damageBaddie(baddie, damage, angle) {
  baddie.health = Math.max(baddie.health - damage, 0);

  play2Dsound(`${baddie.health === 0 ? "deaths" : "hits"}.baddies.${baddie.baddieDesc.id}`, baddie.pos);

  if (baddie.health === 0) {
    killBaddie(baddie);
  } else {
    setBaddieHit(baddie, angle);
  }
}

/**
 * @param {Baddie} baddie
 */
function killBaddie(baddie) {
  baddie.spawner.currentAmount--;

  inGame.activeWorld.pixiLayers["ground"].entitiesContainer.removeChild(baddie.pixiRoot);
  inGame.activeWorld.pixiShadowsLayer.removeChild(baddie.shadowSprite);
  inGame.activeWorld.baddies.splice(inGame.activeWorld.baddies.indexOf(baddie), 1);

  baddie.bodySprite.destroy();
  baddie.shadowSprite.destroy();

  spawnAnimatedFx("inGame/fxs/vanish", baddie.pos, 0);

  if (baddie.baddieDesc.dropItem != null) {
    const item = createItem(inGame.activeWorld, baddie.pos, getItemDesc(baddie.baddieDesc.dropItem));
    inGame.activeWorld.items.push(item);
  }

  if (randInt(0, 3) === 0) {
    const type = randInt(0, 3) === 0 ? "armor" : "heart";
    getSound(`pickUp.${type}`).play();

    if (type === "heart") inGame.stats.health = Math.min(inGame.stats.maxHealth, inGame.stats.health + 1);
    else inGame.stats.armor = Math.min(inGame.stats.maxArmor, inGame.stats.armor + 1);
    updateHudHealth();

    let timer = 0;
    const duration = 0.5;

    spawnFx(`inGame/fxs/${type}PickUp`, baddie.pos, 0, (fx, dt) => {
      timer = Math.min(duration, timer + dt);
      const newPos = lerpxy(fx.sprite.position, inGame.baxton.pos, timer / duration);
      fx.sprite.position.set(newPos.x, newPos.y);
      return timer < duration;
    });
  }
}